from time import clock
import matplotlib.pyplot as pp
from random import randint
from math import ceil

## Exponentiation rapide 1
def Fpow(a,n):
    if n==0:
        return 1
    else:
        p=n//2
        b=Fpow(a,p)
        if n%2==0:
            return b*b
        else:
            return a*b*b
## Complexité temporelle :
# C(n)= 1 + C(n/2) + 2 + 1 ou 2 <= 6+C(n/2)
# Si n=2**p : C(2**p)<= 6+C(2**(p-1)) et par récurrence facile C(2**p)<= 6p .
# Sinon on encadre n entre 2 puissances de 2 : 2**p<n<2**(p+1) et comme la complexité est croissante 3p<C(n)<3(p+1) d'où C(n)=O(p)=O(log2(n)) ce qui est très bon.

## Partages 1
def partage(n,p):
    assert p>=1
    if p==1:
        return [[n]]  # liste constituée de la liste [n]
    else:
        M=[]  # initialisation de la liste des partages
        for i in range(ceil(n/p),n-p+2):  # élément maximal imposé du partage
            L=partage(n-i,p-1)  # liste de listes
            for l in L:  # l est une liste d'entiers décroissants 
                if l[0]<=i:  # test pour voir si l est compatible avec i
                    M.append([i]+l)  # [i]+l est un partage valide commençant par i
                # sinon on ne fait rien car i doit être maximal
    return M
    
## Partages 2
def temps(n,p):
    t0=clock()
    partage(n,p)
    t1=clock()
    return t1-t0

## Partages 3
def grafp(n):
    pp.close()
    X=list(range(1,n+1))
    Y=[temps(n,p) for p in X]
    pp.plot(X,Y)
    pp.show()
    
def grafn(p):
    pp.close()
    X=list(range(p,p+21))
    Y=[temps(n,p) for n in X]
    pp.plot(X,Y)
    pp.show()
    
## Dédoublonnage 1
def sd1(L):
    if len(L)<=1:
        return L
    else:
        d,l=L[-1],L[:-1]  # ceci ne modifie pas L
        l=sd1(l)
        if d not in l:
            l.append(d)
        return l
        
## Complexités 
# C(n)=2+2+C(n-1)+2  donc C(n) est arithmétique donc C(n)=O(n).
# M(n)=n+n+M(n-1)+1 < 3n+M(n-1)  donc M(n)<3(1+2+...+n-1)=3n(n-1)/2=O(n²).
        
## Dédoublonnage 3     
def sd2(L):
    if len(L)>=2:
        d=L.pop()  # ceci modifie L
        sd2(L)
        if d not in L:
            L.append(d)
            
# Même complexité temporelle que sd1 (à peine modifié).
# M(n)=n+1+M(n-1)+1  et ensuite comme pour sd1, on obtient encore O(n²).

## Dédoublonnage 4
def compare(n):
    L=[randint(0,9) for i in range(n)]
    t0=clock()
    sd1(L)  # ne modifie pas  L
    t1=clock()
    sd2(L)
    t2=clock()
    return t1-t0,t2-t1
    
## Dédoublonnage 5
def sd3(L):
    if len(L)==1:
        return [(L[0],1)]
    else:
        d=L.pop()
        L=sd3(L) # renvoie une liste de couples
        NL=[] # initialisation de la nouvelle liste
        b=False # initialisation du booléen qui permettra de savoir si d apparait dans l
        for c in L:
            if d==c[0]: 
                NL.append((d,c[1]+1)) # d apparait une fois de plus qu'avant
                b=True # d était dans l
            else:
                NL.append(c) # on ne modifie pas c
        if not b: # cas où d n'était pas dans l
            NL.append((d,1))
        return NL
        
## Fibonacci 2
dico={0:0,1:1}
# Initialisation du dictionnaire: f(0)=0, f(1)=1 .
def fib(n):
    global dico  # dico doit être visible depuis fib, et récupérable après ;
    if n in dico:
        return dico[n]  # si f(n) est déjà dans le dictionnaire on le lit directement ;
    else:
        q=(n+1)//2
        r=q-1  # si n=2p, q=p ; si n=2p+1, q=p+1 ;
        if q in dico:
            a=dico[q]  # on lit f(q) dans le dictionnaire
        else:
            a=fib(q)  # appel récursif ;
            dico[q]=a  # on complète le dictionnaire ;
        if r in dico:
            b=dico[r]
        else:
            b=fib(r)
            dico[r]=b
    if n%2==1:
        f=a**2+b**2  # formule dans le cas n impair ;
    else:
        f=a*(2*b+a)  # formule dans le cas n pair ;
    dico[n]=f  # on complète le dictionnaire.
    return f
    
## Fibonacci 3
dico={0:0,1:1}
def fibultim(x):
    c=0
    # compteur d'appels récursifs ;
    def fib(n):
        global dico
        nonlocal c  # c doit être visible de fibultim mais n'est pas globale car locale de fibultim ;
        if n in dico:
            return dico[n]
        else:
            q=(n+1)//2
            r=q-1
            if q in dico:
                a=dico[q]
            else:
                a=fib(q)  # appel récursif
                c+=1  # donc on incrémente le compteur ;
                dico[q]=a
            if r in dico:
                b=dico[r]
            else:
                b=fib(r)
                c+=1  # idem
                dico[r]=b
        if n%2==1:
            f=a**2+b**2
        else:
            f=a*(2*b+a)
        dico[n]=f
        return f  # fin de la fonction fib
    y=fib(x)
    print('Le',x,'ème nombre de Fibonacci est',y,'obtenu avec',c,'appels récursifs')
    print('Le dictionnaire est',dico)
    # état du dictionnaire après les derniers calculs.